/*******************************************************************************

NAME            cba_mem.c
VERSION         %version:   1 %
PROGRAMMER      %created_by: mben %

    Copyright 2017 NetApp, Inc.

LICENSE:
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include <config-util.h>
#include <grub/env.h>
#include <grub/file.h>
#include <grub/dl.h>
#include <grub/command.h>

GRUB_MOD_LICENSE ("GPLv3+");
GRUB_MOD_DUAL_LICENSE ("All Rights Reserved");

#define PARAM_FILE  "(lvm/system-cba)/cba.cfg"

#define DBG_PRT(arg...)    \
{                          \
    if (enableDbgPrints)   \
        grub_printf(arg);  \
}

static int enableDbgPrints = 0;

static const char memParamString[]="MEMORY=";

/**
 * This function opens a the CBA parameter file, scans
 * it for the memory request, and returns it if present.
 */
static int
grub_read_mem(char *memsz_str, int memsz_len)
{
    grub_file_t file;
    char*       read_buf;
    char*       read_buf_end;
    char*       buf_ptr;
    grub_off_t  file_len;
    unsigned long mem_req = 0;
    unsigned long mem_param_hdr_size;

    /* Read in file to a temporary buffer */
    file = grub_file_open(PARAM_FILE);
    if (!file)
    {
        DBG_PRT("Could not open file: %s\n", PARAM_FILE);
        return -1;
    }

    read_buf = (char *)grub_zalloc(file->size);
    if (!read_buf)
    {
        grub_printf("Allocation failure %" PRIuGRUB_UINT64_T " bytes)\n", file->size);
        grub_file_close(file);
        return -1;
    }

    file_len = grub_file_read(file, read_buf, file->size);
    if (file_len != file->size)
    {
        grub_printf("Failed to read file (read return 0x%" PRIxGRUB_UINT64_T " on attempted read of %" PRIuGRUB_UINT64_T ")\n", file_len, file->size);
        grub_free(read_buf);
        grub_file_close(file);
        return -1;
    }
    grub_file_close(file);

    mem_param_hdr_size = grub_strlen(memParamString);
    read_buf_end = read_buf + file_len;
    buf_ptr = read_buf;

    while (buf_ptr < read_buf_end)
    {
        /* Skip comments and lines not matching 'MEMORY='. */
        if (*buf_ptr == '#'
            || grub_strncmp(buf_ptr, memParamString, mem_param_hdr_size) != 0)
        {
            /* Advance to the next new line. */
            while (buf_ptr < read_buf_end && *buf_ptr != '\n')
                ++buf_ptr;

            /* Consume the new line. */
            ++buf_ptr;

            continue;
        }

        DBG_PRT("Found header.");
        buf_ptr += mem_param_hdr_size;
        mem_req = grub_strtoul(buf_ptr, 0, 10);
        break;
    }

    /* Reset buf_ptr to start of file data. */
    buf_ptr = read_buf;

    /* If the string parsing didn't find a valid number above, the
     * strtoul conversion would return a 0, which is what we would
     * want by default.
     */
    DBG_PRT("Requested memory size: %lu.", mem_req);
    grub_snprintf(memsz_str, memsz_len, "%lu", mem_req);
    grub_free(read_buf);
    return 0;
}


/**
 * This function implements the cba_mem grub command.
 * It scans the CBA parameters file for the requested memory and
 * creates/sets the grub environment variable with the value. If
 * the parameter file does not exist, it will set the variable to
 * 0, assuming no CBA memory is required.
 */
static grub_err_t
grub_cmd_cba_mem
    (
    grub_command_t cmd __attribute__ ((unused)),
    int argc,
    char *argv[]
    )
{
    char value[128];
    char *varName;
    const char *defaultValue = "0";

    if (argc != 1)
    {
        return grub_error(GRUB_ERR_BAD_ARGUMENT, N_("one arguments expected"));
    }

    /* If "DEBUG" env var exists, enable prints */
    enableDbgPrints = grub_env_get("DEBUG") ? 1 : 0;

    varName = argv[0];
    DBG_PRT("variable: %s\n", varName);

    if (grub_read_mem(value, sizeof(value)))
    {
        /* If it failed to read the value from config file, then assign the default value */
        grub_strncpy(value, defaultValue, sizeof(value));

        /* Ignore the error to avoid waiting for user response during boot */
        grub_errno = 0;
    }

    DBG_PRT("Set %s=%s", varName, value);
    grub_env_set(varName, value);

    return GRUB_ERR_NONE;
}

static grub_command_t cmd;

/* Register/unregister our command with module init/exit */
GRUB_MOD_INIT(cba_mem)
{
    cmd = grub_register_command("cba_mem", grub_cmd_cba_mem, N_("ENV_VAR"),
                                N_("Provides memory in GB requested for CBA."));
}

GRUB_MOD_FINI(cba_mem)
{
    if (cmd)
    {
        grub_unregister_command(cmd);
        cmd = 0;
    }
}
